/*
 * drivers/input/misc/ep93xx-sw.c
 * Input switch driver for ep93xx
 *
 * Copyright (C) 2009 Atmark Techno, Inc. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/input.h>
#include <linux/irq.h>
#include <linux/gpio.h>

extern int ep93xx_gpio_debounce_set(unsigned int gpio, int arg);

struct ep93xx_sw_info {
	struct input_dev *idev;
	struct ep93xx_gpio_pin *pin;
};

static irqreturn_t ep93xx_sw_irq_handler(int irq, void *dev_id)
{
	struct input_dev *idev = (struct input_dev *) dev_id;
	unsigned int gpio = irq_to_gpio(irq);

	if (gpio_get_value(gpio)) {
		set_irq_type(irq, IRQT_FALLING);
		input_event(idev, EV_SW, 0, 0);	/* Up! */
	} else {
		set_irq_type(irq, IRQT_RISING);
		input_event(idev, EV_SW, 0, 1);	/* Down! */
	}

	input_sync(idev);

	return IRQ_HANDLED;
}

static int ep93xx_sw_open(struct input_dev *idev)
{
	struct ep93xx_sw_info *info = dev_get_drvdata(&idev->dev);
	struct ep93xx_gpio_pin *pin = info->pin;
	int irq = gpio_to_irq(pin->gpio);

	if (irq < 0) {
		pr_err("ep93xx-sw: Unable to get irq number"
		       " for GPIO %d, error %d\n", pin->gpio, irq);
		return -ENODEV;
	}

	if (gpio_get_value(pin->gpio))
		set_irq_type(irq, IRQT_FALLING);
	else
		set_irq_type(irq, IRQT_RISING);

	enable_irq(irq);

	return 0;
}

static void ep93xx_sw_close(struct input_dev *idev)
{
	struct ep93xx_sw_info *info = dev_get_drvdata(&idev->dev);
	struct ep93xx_gpio_pin *pin = info->pin;
	int irq = gpio_to_irq(pin->gpio);

	if (irq < 0) {
		pr_err("ep93xx-sw: Unable to get irq number"
		       " for GPIO %d, error %d\n", pin->gpio, irq);
		return;
	}

	disable_irq(irq);
}

static void ep93xx_sw_set_input(struct input_dev *input,
				struct platform_device *pdev,
				struct ep93xx_gpio_pin *pin)
{
	input->name = pin->name;
	input->phys = NULL;
	input->id.bustype = BUS_HOST;
	input->dev.parent = &pdev->dev;

	input->open = ep93xx_sw_open;
	input->close = ep93xx_sw_close;

	input_set_capability(input, EV_SW, 0);
}

static int ep93xx_sw_set_gpio(struct ep93xx_gpio_pin *pin,
			      struct input_dev *input)
{
	int error;
	int irq;

	error = gpio_request(pin->gpio, pin->name);
	if (error) {
		pr_err("ep93xx-sw: failed to request GPIO %d, error %d\n",
		       pin->gpio, error);
		return error;
	}

	gpio_direction_input(pin->gpio);
	ep93xx_gpio_debounce_set(pin->gpio, 1);

	irq = gpio_to_irq(pin->gpio);
	if (irq < 0) {
		pr_err("ep93xx-sw: Unable to get irq number"
		       " for GPIO %d, error %d\n", pin->gpio, error);
		gpio_free(pin->gpio);
		return irq;
	}

	error = request_irq(irq,
			    ep93xx_sw_irq_handler,
			    IRQF_DISABLED, pin->name, input);

	if (error) {
		pr_err("ep93xx-sw: Unable to claim irq %d; error %d\n",
		       irq, error);
		gpio_free(pin->gpio);
		return error;
	}

	disable_irq(irq);

	return 0;
}

static int ep93xx_sw_probe(struct platform_device *pdev)
{
	struct ep93xx_gpio_platform_data *plat_data = pdev->dev.platform_data;
	struct ep93xx_sw_info *info;
	int error;
	int i;

	info = kzalloc(sizeof(struct ep93xx_sw_info) *
		       plat_data->num_gpios, GFP_KERNEL);
	if (!info)
		return -ENOMEM;

	for (i = 0; i < plat_data->num_gpios; i++) {
		struct ep93xx_gpio_pin *pin = &plat_data->gpios[i];
		struct input_dev *input = input_allocate_device();
		if (!input) {
			error = -ENOMEM;
			pr_err("ep93xx-sw: failed to allocate input device %d, error %d\n",
			       i, error);
			goto fail;
		}

		ep93xx_sw_set_input(input, pdev, pin);

		error = input_register_device(input);
		if (error) {
			input_free_device(input);
			pr_err("ep93xx-sw: failed to register input device %d, error %d\n",
			       i, error);
			goto fail;
		}

		error = ep93xx_sw_set_gpio(pin, input);
		if (error) {
			input_unregister_device(input);
			goto fail;
		}

		info[i].pin = pin;
		info[i].idev = input;
		dev_set_drvdata(&input->dev, &info[i]);
	}

	platform_set_drvdata(pdev, info);

	return 0;

 fail:
	platform_set_drvdata(pdev, NULL);

	while (--i >= 0) {
		input_unregister_device(info[i].idev);
		free_irq(gpio_to_irq(info[i].pin->gpio), info[i].idev);
		gpio_free(info[i].pin->gpio);
	}

	kfree(info);

	return error;
}

static int ep93xx_sw_remove(struct platform_device *pdev)
{
	struct ep93xx_gpio_platform_data *plat_data = pdev->dev.platform_data;
	struct ep93xx_sw_info *info = platform_get_drvdata(pdev);
	int i;

	platform_set_drvdata(pdev, NULL);

	for (i = 0; i < plat_data->num_gpios; i++) {
		input_unregister_device(info[i].idev);
		free_irq(gpio_to_irq(info[i].pin->gpio), info[i].idev);
		gpio_free(info[i].pin->gpio);
	}

	kfree(info);

	return 0;
}

static struct platform_driver ep93xx_sw_driver = {
	.probe = ep93xx_sw_probe,
	.remove = __devexit_p(ep93xx_sw_remove),
	.driver = {
		.name = "ep93xx-sw",
	},
};

static int __init ep93xx_sw_init(void)
{
	return platform_driver_register(&ep93xx_sw_driver);
}

static void __exit ep93xx_sw_exit(void)
{
	platform_driver_unregister(&ep93xx_sw_driver);
}

module_init(ep93xx_sw_init);
module_exit(ep93xx_sw_exit);

MODULE_AUTHOR("Atmark Techno, Inc.");
MODULE_DESCRIPTION("Input switch driver for ep93xx");
MODULE_LICENSE("GPL v2");
