/*
 * drivers/mtd/nand/ep93xx.c
 * NAND driver for ep93xx
 *
 * Copyright (C) 2009 Atmark Techno, Inc. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/gpio.h>
#include <linux/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/partitions.h>

#include <asm/arch/ep93xx-regs.h>
#include <asm/arch/mtd.h>

struct ep93xx_nand_mtd {
	struct mtd_info             mtd_info;
	struct nand_chip            nand_chip;
	struct ep93xx_nand_platdata plat;
};

#define ep93xx_nand_getpriv(x) container_of(x, struct ep93xx_nand_mtd, mtd_info)

static void ep93xx_nand_cmd_ctrl(struct mtd_info *mtd_info, int cmd, unsigned int ctrl)
{
	struct ep93xx_nand_mtd *mtd = ep93xx_nand_getpriv(mtd_info);

	if (ctrl & NAND_CTRL_CHANGE) {
		gpio_set_value(mtd->plat.gpio_nce, !(ctrl & NAND_NCE));
		gpio_set_value(mtd->plat.gpio_cle, !!(ctrl & NAND_CLE));
		gpio_set_value(mtd->plat.gpio_ale, !!(ctrl & NAND_ALE));
	}
	if (cmd == NAND_CMD_NONE)
		return;

	writeb(cmd, mtd->nand_chip.IO_ADDR_W);
}

static int ep93xx_nand_dev_ready(struct mtd_info *mtd_info)
{
	struct ep93xx_nand_mtd *mtd = ep93xx_nand_getpriv(mtd_info);
	return gpio_get_value(mtd->plat.gpio_rdy);
}

static void __iomem *request_and_remap(struct resource *res, size_t size,
					const char *name, int *err)
{
	void __iomem *ptr;

	if (!request_mem_region(res->start, res->end - res->start + 1, name)) {
		*err = -EBUSY;
		return NULL;
	}

	ptr = ioremap(res->start, size);
	if (!ptr) {
		release_mem_region(res->start, res->end - res->start + 1);
		*err = -ENOMEM;
	}
	return ptr;
}

static int ep93xx_gpio_request_input(unsigned int gpio, const char *name)
{
	int ret;

	ret = gpio_request(gpio, name);
	if (ret)
		return ret;

	gpio_direction_input(gpio);

	return 0;
}

static int ep93xx_gpio_request_output(unsigned int gpio, const char *name, int val)
{
	int ret;

	ret = gpio_request(gpio, name);
	if (ret)
		return ret;

	gpio_direction_output(gpio, val);

	return 0;
}

static inline void ep93xx_gpio_free(unsigned int gpio)
{
	gpio_free(gpio);
}

static void ep93xx_nand_set_syscon(void)
{
	unsigned long temp;

	temp = inl(EP93XX_SYSCON_DEVICE_CONFIG);
	outl(0xaa, EP93XX_SYSCON_SWLOCK);                        /* Unlock Software Lock */
	outl(temp | 0x00000002, EP93XX_SYSCON_DEVICE_CONFIG);    /* Key Matrix Controller inactive */
	outl(0x000004c9, EP93XX_PCMCIA_CONTROLLER_BASE + 0x0c);  /* SMCBCR3 */
}

static void register_mtd_part(struct ep93xx_nand_mtd *mtd)
{
#ifdef CONFIG_MTD_PARTITIONS
	const char *part_probes[] = {"cmdlinepart", NULL};
	struct mtd_partition *mtd_parts;
	int nr_parts = 0;

	pr_info("%s\n", __func__);

	mtd->mtd_info.name = mtd->plat.name;

	nr_parts =
	    parse_mtd_partitions(&mtd->mtd_info, part_probes, &mtd_parts, 0);

	if (nr_parts > 0) {
		pr_info("Registering %s using command line partitions\n", mtd->mtd_info.name);
		add_mtd_partitions(&mtd->mtd_info, mtd_parts, nr_parts);
		mtd->plat.parts = mtd_parts;
		mtd->plat.num_parts = nr_parts;
	} else if (mtd->plat.parts) {
		pr_info("Registering %s using default partitions\n", mtd->mtd_info.name);
		add_mtd_partitions(&mtd->mtd_info, mtd->plat.parts, mtd->plat.num_parts);
	} else
#endif
	{
		pr_info("Registering %s as whole device\n", mtd->mtd_info.name);
		add_mtd_device(&mtd->mtd_info);
	}
}

static int __devinit ep93xx_nand_probe(struct platform_device *dev)
{
	struct ep93xx_nand_mtd *mtd;
	struct nand_chip *this;
	struct resource *res;
	int ret = 0;

	if (dev->dev.platform_data == NULL)
		return -EINVAL;

	res = platform_get_resource(dev, IORESOURCE_MEM, 0);
	if (res == NULL)
		return -EINVAL;

	mtd = kzalloc(sizeof(*mtd), GFP_KERNEL);
	if (mtd == NULL) {
		dev_err(&dev->dev, "failed to create NAND MTD\n");
		return -ENOMEM;
	}

	this = &mtd->nand_chip;
	this->IO_ADDR_R = request_and_remap(res, 2, "NAND", &ret);
	if (this->IO_ADDR_R == NULL) {
		dev_err(&dev->dev, "unable to map NAND\n");
		goto err_map;
	}

	memcpy(&mtd->plat, dev->dev.platform_data, sizeof(mtd->plat));

	ret = ep93xx_gpio_request_output(mtd->plat.gpio_nce, "NAND NCE", 1);
	if (ret)
		goto err_nce;
	ret = ep93xx_gpio_request_output(mtd->plat.gpio_ale, "NAND ALE", 0);
	if (ret)
		goto err_ale;
	ret = ep93xx_gpio_request_output(mtd->plat.gpio_cle, "NAND CLE", 0);
	if (ret)
		goto err_cle;
	if (gpio_is_valid(mtd->plat.gpio_rdy)) {
		ret = ep93xx_gpio_request_input(mtd->plat.gpio_rdy, "NAND RDY");
		if (ret)
			goto err_rdy;
	}

	this->IO_ADDR_W  = this->IO_ADDR_R;
	this->ecc.mode   = NAND_ECC_SOFT;
	this->options    = mtd->plat.options;
	this->chip_delay = mtd->plat.chip_delay;

	this->cmd_ctrl   = ep93xx_nand_cmd_ctrl;
	if (gpio_is_valid(mtd->plat.gpio_rdy))
		this->dev_ready  = ep93xx_nand_dev_ready;
	else
		this->dev_ready  = NULL;

	mtd->mtd_info.priv  = this;
	mtd->mtd_info.owner = THIS_MODULE;

	ep93xx_nand_set_syscon();

	if (nand_scan(&mtd->mtd_info, 1)) {
		dev_err(&dev->dev, "no nand chips found\n");
		ret = -ENXIO;
		goto err_nand_scan;
	}

	register_mtd_part(mtd);

	platform_set_drvdata(dev, mtd);

	return 0;

err_nand_scan:
	if (gpio_is_valid(mtd->plat.gpio_rdy))
		ep93xx_gpio_free(mtd->plat.gpio_rdy);
err_rdy:
	ep93xx_gpio_free(mtd->plat.gpio_cle);
err_cle:
	ep93xx_gpio_free(mtd->plat.gpio_ale);
err_ale:
	ep93xx_gpio_free(mtd->plat.gpio_nce);
err_nce:
	iounmap(mtd->nand_chip.IO_ADDR_R);
	release_mem_region(res->start, res->end - res->start + 1);
err_map:
	kfree(mtd);

	return ret;
}

static int __devexit ep93xx_nand_remove(struct platform_device *dev)
{
	struct ep93xx_nand_mtd *mtd = platform_get_drvdata(dev);
	struct resource *res;

	nand_release(&mtd->mtd_info);

	res = platform_get_resource(dev, IORESOURCE_MEM, 0);
	iounmap(mtd->nand_chip.IO_ADDR_R);
	release_mem_region(res->start, res->end - res->start + 1);

	gpio_set_value(mtd->plat.gpio_nce, 1);

	ep93xx_gpio_free(mtd->plat.gpio_cle);
	ep93xx_gpio_free(mtd->plat.gpio_ale);
	ep93xx_gpio_free(mtd->plat.gpio_nce);

	if (gpio_is_valid(mtd->plat.gpio_rdy))
		ep93xx_gpio_free(mtd->plat.gpio_rdy);

	kfree(mtd);

	return 0;
}

static struct platform_driver ep93xx_nand_driver = {
	.probe        = ep93xx_nand_probe,
	.remove       = ep93xx_nand_remove,
	.driver       = {
		.name = "ep93xx-nand"
	},
};

static int __init ep93xx_nand_init(void)
{
	pr_info("NAND Driver for ep93xx\n");

	return platform_driver_register(&ep93xx_nand_driver);

}

static void __exit ep93xx_nand_exit(void)
{
	platform_driver_unregister(&ep93xx_nand_driver);
}

module_init(ep93xx_nand_init);
module_exit(ep93xx_nand_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Atmark Techno, Inc.");
MODULE_DESCRIPTION("NAND driver for ep93xx");
