/*
 * drivers/pcmcia/ep93xx_pcmcia.c
 *
 * Copyright 2010 Atmark Techno, Inc. All Rights Reserved.
 *
 * Based on the Cirrus Logic ep93xx_pcmcia driver.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/autoconf.h>
#include <linux/cpufreq.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/platform_device.h>
#include <linux/irq.h>
#include <linux/gpio.h>
#include <linux/clk.h>

#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/system.h>
#include <asm/arch/pcmcia.h>

#include <pcmcia/cs_types.h>
#include <pcmcia/ss.h>
#include <pcmcia/bulkmem.h>
#include <pcmcia/cistpl.h>

#define DRIVER_NAME "ep93xx-pcmcia"
#define AUTHOR      "Atmark Techno, Inc."
#define DESCRIPTION "PCMCIA driver"

/* Maximum number of IO windows per socket */
#define EP93XX_MAX_IO_WIN 2

/* Maximum number of memory windows per socket */
#define EP93XX_MAX_MEM_WIN 5

#define EP93XX_UNCONFIGURED_VOLTAGE 0xFF

struct pcmcia_irq_info {
	unsigned int sock;
	unsigned int irq;
};

static struct irqs {
	int irq;		/* Interrupt number */
	unsigned int gpio;	/* GPIO port F pin mask */
	const char *str;	/* String */
} ep93xx_pcmcia_irqs[] = {
	{EP93XX_PCMCIA_INT_READY,  EP93XX_PCMCIA_READY,  DRIVER_NAME " ready"},
	{EP93XX_PCMCIA_INT_MCD1,   EP93XX_PCMCIA_MCD1,   DRIVER_NAME " cd1"},
	{EP93XX_PCMCIA_INT_MCD2,   EP93XX_PCMCIA_MCD2,   DRIVER_NAME " cd2"},
	{EP93XX_PCMCIA_INT_MCBVD2, EP93XX_PCMCIA_MCBVD2, DRIVER_NAME " mcbvd2"},
	{EP93XX_PCMCIA_INT_MCBVD1, EP93XX_PCMCIA_MCBVD1, DRIVER_NAME " mcbvd1"},
};

#define EP93XX_GPIOS								\
	(EP93XX_PCMCIA_READY |	EP93XX_PCMCIA_MCD1 | EP93XX_PCMCIA_MCD2 |	\
	 EP93XX_PCMCIA_MCBVD2 | EP93XX_PCMCIA_MCBVD1)


/* This structure maintains housekeeping state for each socket, such
 * as the last known values of the card detect pins, or the Card Services
 * callback value associated with the socket:
 */
static struct socket_info_t {
	struct pcmcia_socket	socket;
	struct device		*dev;
	/*the number of the sockets*/
	unsigned int		nr;
	unsigned int		irq;
	struct list_head	node;

	socket_state_t		state;
	struct pccard_mem_map 	mem_map[EP93XX_MAX_MEM_WIN];
	struct pccard_io_map	io_map[EP93XX_MAX_IO_WIN];

	void			*virt_io;
	unsigned short		speed_io, speed_attr, speed_mem;
	struct resource		*res_io, *res_attr, *res_mem;
	unsigned int		status;
	unsigned long		enable_irq;
	u_char			vcc;

	struct clk *hclk;
} socket_info;

struct bittbl {
	unsigned int mask;
	const char *name;
};

static struct bittbl status_bits[] = {
	{SS_WRPROT, "SS_WRPROT"},
	{SS_BATDEAD, "SS_BATDEAD"},
	{SS_BATWARN, "SS_BATWARN"},
	{SS_READY, "SS_READY"},
	{SS_DETECT, "SS_DETECT"},
	{SS_POWERON, "SS_POWERON"},
	{SS_STSCHG, "SS_STSCHG"},
	{SS_3VCARD, "SS_3VCARD"},
	{SS_XVCARD, "SS_XVCARD"},
};

static struct bittbl conf_bits[] = {
	{SS_PWR_AUTO, "SS_PWR_AUTO"},
	{SS_IOCARD, "SS_IOCARD"},
	{SS_RESET, "SS_RESET"},
	{SS_DMA_MODE, "SS_DMA_MODE"},
	{SS_SPKR_ENA, "SS_SPKR_ENA"},
	{SS_OUTPUT_ENA, "SS_OUTPUT_ENA"},
};


static void dump_bits(char **p, const char *prefix, unsigned int val,
		      struct bittbl *bits, int sz)
{
	char *b = *p;
	int i;

	b += sprintf(b, "%-9s:", prefix);
	for (i = 0; i < sz; i++)
		if (val & bits[i].mask)
			b += sprintf(b, " %s", bits[i].name);
	*b++ = '\n';
	*p = b;
}

/*
 * Implements the /sys/class/pcmcia_socket/??/status file.
 *
 * Returns: the number of characters added to the buffer
 */
static ssize_t show_status(struct device *dev,
			   struct device_attribute *attr, char *buf)
{
	struct socket_info_t *skt = &socket_info;
	char *p = buf;

	p += sprintf(p, "slot     : %d\n", skt->nr);

	dump_bits(&p, "status", skt->status,
		  status_bits, ARRAY_SIZE(status_bits));
	dump_bits(&p, "csc_mask", skt->state.csc_mask,
		  status_bits, ARRAY_SIZE(status_bits));
	dump_bits(&p, "cs_flags", skt->state.flags,
		  conf_bits, ARRAY_SIZE(conf_bits));

	p += sprintf(p, "Vcc      : %d\n", skt->state.Vcc);
	p += sprintf(p, "Vpp      : %d\n", skt->state.Vpp);
	p += sprintf(p, "IRQ      : %d (%d)\n", skt->state.io_irq, skt->irq);

	return p - buf;
}
static DEVICE_ATTR(status, S_IRUGO, show_status, NULL);

static void ep93xx_pcmcia_power_control(struct pcmcia_socket *sock, int mode)
{
	int timeout = 300;
	u32 status;

	if (mode) { /* Power ON */
		do {
			gpio_direction_output(EP93XX_PCMCIA_POWER, 1);

			sock->ops->get_status(sock, &status);
			if (status & SS_POWERON)
				break;
			if (timeout-- < 0) {
				pr_debug(DRIVER_NAME ": unable to activate "
					 "socket power\n");
				break;
			}
			mdelay(1);
		} while (1);
	} else { /* Power OFF */
		do {
			gpio_direction_output(EP93XX_PCMCIA_POWER, 0);

			sock->ops->get_status(sock, &status);
			if (!(status & SS_POWERON))
				break;
			if (timeout-- < 0) {
				pr_debug(DRIVER_NAME ": unable to remove "
					 "socket power\n");
				break;
			}
			mdelay(1);
		} while (1);
	}
}

static inline void ep93xx_pcmcia_power_on(struct pcmcia_socket *sock)
{
	pr_debug(DRIVER_NAME ": ep93xx_pcmcia_power_on\n");

	ep93xx_pcmcia_power_control(sock, 1);
}

static inline void ep93xx_pcmcia_power_off(struct pcmcia_socket *sock)
{
	pr_debug(DRIVER_NAME ": ep93xx_pcmcia_power_off\n");

	ep93xx_pcmcia_power_control(sock, 0);
}

static int ep93xx_set_voltage(u_char vcc)
{
	struct socket_info_t *skt = &socket_info;
	unsigned long switch_setting;

	pr_debug(DRIVER_NAME ": ep93xx_set_voltage\n");

	if (skt->vcc == vcc) {
		pr_debug(DRIVER_NAME ": Power already set to %d\n", vcc );
		return 0;
	}
	
	switch_setting = EE_ADDRESS | ENABLE;
	switch (vcc) {
	case 0:
		pr_debug(DRIVER_NAME ": Configure the socket for 0 Volts\n");
		switch_setting |= AVCC_0V;
		break;
	case 50:
		pr_debug(DRIVER_NAME ": Configure the socket for 5 Volts\n");
		switch_setting |= AVCC_5V;
		break;
	case 33:
		pr_debug(DRIVER_NAME ": Configure the socket for 3.3 Volts\n");
		switch_setting |= AVCC_33V;
		break;
	default:
		pr_err(DRIVER_NAME ": %s(): unrecognized Vcc %u\n",
		       __FUNCTION__, vcc);
		return -1;
	}

	if (switch_setting & AVCC_33V)
		ep93xx_pcmcia_power_off((struct pcmcia_socket *)&(skt->socket));
	else
		ep93xx_pcmcia_power_on((struct pcmcia_socket *)&(skt->socket));


	skt->vcc = vcc;

	return 0;
}

#define PCMCIA_BOARD_DELAY 40
/*
 * Calculate the pcmcia timings based on the register settings. 
 * For example here is for Attribute/Memory Read.
 *
 *
 * Address:   _______XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX___________
 *
 * Data:      __________________________XXXXXXXXXXXXXX________________________
 *
 * CE#:       -------______________________________________________-----------
 *
 * REG#:      -------______________________________________________-----------
 *
 * OE#:       --------------------___________________-------------------------
 *
 *
 *                   |<---------------Cycle time ----------------->|
 *
 *                   |< Address >|<-- Access Time -->|< Hold Time >|
 *                      Time           ta(CE)             th(CE)
 *                      tsu(A)
 *
 * See PCMCIA Electrical Specification Section 4.7 for the timing numbers.
 *
 */
static unsigned long ep93xx_calculate_timing(unsigned long speed)
{
	struct socket_info_t *skt = &socket_info;
	unsigned long addr_time, hold_time, access_time;
	unsigned long access_cycle, address_cycle, hold_cycle;
	unsigned long period, smc;
	unsigned long hclk = clk_get_rate(skt->hclk);

	switch (speed) {
	/* Special case for I/O all access. */
	case 0:
		access_time = 165;
		addr_time = 70;
		hold_time = 20;
		break;
	case 100:
		access_time = 100;
		addr_time = 10;
		hold_time = 15;
		break;
	case 150:
		access_time = 150;
		addr_time = 20;
		hold_time = 20;
		break;
	case 200:
		access_time = 200;
		addr_time = 20;
		hold_time = 20;
		break;
	case 250:
		access_time = 250;
		addr_time = 30;
		hold_time = 20;
		break;
	case 300:
		access_time = 300;
		addr_time = 30;
		hold_time = 20;
		break;
	case 600:
	default:
		access_time = 600;
		addr_time = 100;
		hold_time = 35;
		break;
	}

	/* Add in a board delay. */
	access_time += PCMCIA_BOARD_DELAY;
	addr_time   += PCMCIA_BOARD_DELAY;
	hold_time   += PCMCIA_BOARD_DELAY;
	
	/*
	 * This gives us the period in nanosecods.
	 *
	 * = 1000000000(ns/s) / HCLK(cycle/s)
	 * = (ns/cycle)
	 */
	period = 1000000000 / hclk;
	
	/*
	 * Find the number of hclk cycles for cycle time, address time and
	 * hold time.
	 *
	 * = access_time(ns) / period(ns/Cycles)
	 * = addr_time(ns)   / period(ns/Cycles)
	 * = hold_time(ns)   / period(ns/Cycles)
	 */
	access_cycle = access_time / period;
	if (access_cycle > 0xFF)
		access_cycle = 0xFF;
	
	address_cycle = (addr_time / period) + 1;
	if (address_cycle > 0xFF)
		address_cycle = 0xFF;
			
	hold_cycle = (hold_time / period) + 1;
	if (hold_cycle > 0xF)
		hold_cycle = 0xF;

	smc = (EP93XX_PCMCIA_CONTROLLER_SETUP_TIME_MASK &
	       (address_cycle << EP93XX_PCMCIA_CONTROLLER_SETUP_TIME_SHIFT)) |
		(EP93XX_PCMCIA_CONTROLLER_HOLD_TIME_MASK &
		 (hold_cycle << EP93XX_PCMCIA_CONTROLLER_HOLD_TIME_SHIFT)) |
		(EP93XX_PCMCIA_CONTROLLER_ACCESS_TIME_MASK &
		 (access_cycle << EP93XX_PCMCIA_CONTROLLER_ACCESS_TIME_SHIFT));

	smc |= EP93XX_PCMCIA_CONTROLLER_MEMORY_WIDTH_16BIT;

	return smc;
}

static int ep93xx_get_status(struct pcmcia_socket *sock, unsigned int *status)
{
	struct socket_info_t *skt = &socket_info;
	unsigned long pfdr;

	pr_debug(DRIVER_NAME ": ep93xx_get_status %x\n",sock->sock);

	/* Read the GPIOs that are connected to the PCMCIA slot. */
	pfdr = inl(EP93XX_GPIO_F_DATA);
	
	/* If both CD1 and CD2 are low, set SS_DETECT bit. */
	*status = ((pfdr & (EP93XX_PCMCIA_MCD1 | EP93XX_PCMCIA_MCD2)) == 0) ?
		SS_DETECT : 0;

	*status |= (pfdr & EP93XX_PCMCIA_WP) ? SS_WRPROT : 0;

	if (skt->state.flags & SS_IOCARD)
		*status |= (pfdr & EP93XX_PCMCIA_MCBVD1) ? SS_STSCHG : 0;
	else {
		*status |= (pfdr & EP93XX_PCMCIA_READY) ? SS_READY : 0;
		*status |= (pfdr & EP93XX_PCMCIA_MCBVD1) ? 0 : SS_BATDEAD;
		*status |= (pfdr & EP93XX_PCMCIA_MCBVD2) ? 0 : SS_BATWARN;
	}
	
	*status |= skt->state.Vcc ? SS_POWERON : 0;
	*status |= (pfdr & EP93XX_PCMCIA_VS1) ? 0 : SS_3VCARD;

	return 0;
}

static irqreturn_t ep93xx_interrupt(int irq, void *data)
{
	struct socket_info_t *skt = &socket_info;
	unsigned int status, event;

	/*
	 * We're going to report only the events that have changed and that
	 * are not masked off.
	 */
	ep93xx_get_status(&(skt->socket), &status);
	event = (status ^ skt->status) & skt->state.csc_mask;
	skt->status = status;
	
	if (event)
		pcmcia_parse_events(&skt->socket, event);

	return IRQ_HANDLED;
}

static void ep93xx_enable_irq(unsigned int gpio)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(ep93xx_pcmcia_irqs) ; i++)
		if (ep93xx_pcmcia_irqs[i].gpio & gpio)
			enable_irq(ep93xx_pcmcia_irqs[i].irq);
}

static void ep93xx_disable_irq(unsigned int gpio)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(ep93xx_pcmcia_irqs) ; i++)
		if (ep93xx_pcmcia_irqs[i].gpio & gpio)
			disable_irq(ep93xx_pcmcia_irqs[i].irq);
}

static int ep93xx_set_socket(struct pcmcia_socket *sock, socket_state_t *state)
{
	struct socket_info_t *skt = &socket_info;
	unsigned long tmp;
	unsigned long enable_irq = 0;

	pr_debug(DRIVER_NAME ": ep93xx_set_socket\n");

	/* Set Vcc level. If an illegal voltage is specified, bail w/ error. */
	if (ep93xx_set_voltage(state->Vcc) < 0)
		return -EINVAL;

	/* Enable PCMCIA, Enable Wait States, Set or Clear card reset. */
	tmp = inl(EP93XX_PCMCIA_CONTROLLER_PCMCIA_CTRL) &
		~EP93XX_PCMCIA_CONTROLLER_PCMCIA_CTRL_PCRST;
	tmp |= EP93XX_PCMCIA_CONTROLLER_PCMCIA_CTRL_WEN |
		EP93XX_PCMCIA_CONTROLLER_PCMCIA_CTRL_PCEN;
	if (state->flags & SS_RESET)
		tmp |= EP93XX_PCMCIA_CONTROLLER_PCMCIA_CTRL_PCRST;

	outl(tmp, EP93XX_PCMCIA_CONTROLLER_PCMCIA_CTRL);
	
	/* Enable interrupts in hw that are specified in csc_mask. */
	if (state->csc_mask & SS_DETECT)
		enable_irq |= EP93XX_PCMCIA_MCD1 | EP93XX_PCMCIA_MCD2;

	if (state->flags & SS_IOCARD) {
		if (state->csc_mask & SS_STSCHG)
			enable_irq |= EP93XX_PCMCIA_MCBVD1;
	} else {
		if (state->csc_mask & SS_BATDEAD)
			enable_irq |= EP93XX_PCMCIA_MCBVD1;
		if (state->csc_mask & SS_BATWARN)
			enable_irq |= EP93XX_PCMCIA_MCBVD2;
		if (state->csc_mask & SS_READY)
			enable_irq |= EP93XX_PCMCIA_READY;
	}
	
	skt->enable_irq = enable_irq;

	ep93xx_disable_irq(EP93XX_PCMCIA_READY | enable_irq);
	ep93xx_enable_irq(EP93XX_PCMCIA_READY | enable_irq);

	skt->state = *state;

	return 0;
}


static int ep93xx_set_io_map(struct pcmcia_socket *sock, struct pccard_io_map *io_map)
{
	struct socket_info_t *skt = &socket_info;
	unsigned long pcio;
	unsigned int speed = 0;
	int i;

	pr_debug(DRIVER_NAME ": ep93xx_set_io_map\n");

	if ((io_map->map >= EP93XX_MAX_IO_WIN) || (io_map->stop < io_map->start)) {
		pr_err(DRIVER_NAME ": io_map (%d) out of range\n", io_map->map);
		return -EINVAL;
	}


	if (io_map->flags & MAP_ACTIVE) {
		if (io_map->speed == 0)
			io_map->speed = EP93XX_PCMCIA_IO_ACCESS;

		speed = io_map->speed;

		/*
		 * Go thru our array of mappings, find the lowest speed 
		 * (largest # of nSec) for an active mapping...
		 */
		for (i = 0; i < EP93XX_MAX_IO_WIN; i++) {
			if ((skt->io_map[i].flags & MAP_ACTIVE) && 
			    (speed > skt->io_map[i].speed))
				speed = skt->io_map[i].speed;
		}

		pcio = ep93xx_calculate_timing(speed);
		pcio |= EP93XX_PCMCIA_CONTROLLER_MEMORY_WIDTH_16BIT;

		outl(pcio, EP93XX_PCMCIA_CONTROLLER_PCIO);

		skt->speed_io = speed;
	}

	if (io_map->stop == 1)
		io_map->stop = skt->res_io->end - skt->res_io->start;

	io_map->stop = (io_map->stop - io_map->start) + (unsigned int)skt->virt_io;
	io_map->start = (unsigned int)skt->virt_io;

	skt->io_map[io_map->map] = *io_map;

	return 0;
}



static int ep93xx_set_mem_map(struct pcmcia_socket *sock, struct pccard_mem_map *map)
{
	struct socket_info_t *skt = &socket_info;
	unsigned long reg;
	unsigned int speed = 0;
	int i;


	pr_debug(DRIVER_NAME ": ep93xx_set_mem_map\n");

	if (map->map >= EP93XX_MAX_MEM_WIN) {
		pr_err(DRIVER_NAME ": mem_map (%d) out of range\n", map->map);
		return -EINVAL;
	}

	if (map->flags & MAP_ACTIVE) {
		if (map->speed == 0) {
			if (skt->state.Vcc == 33)
				map->speed = EP93XX_PCMCIA_3V_MEM_ACCESS;
			else
				map->speed = EP93XX_PCMCIA_5V_MEM_ACCESS;
		}
		
		speed = map->speed;

		if (map->flags & MAP_ATTRIB) {
			/*
			 * Go thru our array of mappings, find the lowest speed 
			 * (largest # of nSec) for an active mapping...
			 */
			for (i = 0; i < EP93XX_MAX_MEM_WIN; i++) {
				if ((skt->mem_map[i].flags & MAP_ACTIVE) && 
				    (speed > skt->mem_map[i].speed))
					speed = skt->mem_map[i].speed;
			}

			reg = ep93xx_calculate_timing(speed);
				reg |= EP93XX_PCMCIA_CONTROLLER_MEMORY_WIDTH_16BIT;

			outl(reg, EP93XX_PCMCIA_CONTROLLER_PCATTRIBUTE); 
			
			skt->speed_attr = speed;
			for (i = 0; i < EP93XX_MAX_MEM_WIN; i++) {
				if ((skt->mem_map[i].flags & MAP_ACTIVE) && 
				    (speed > skt->mem_map[i].speed) )
					speed = skt->mem_map[i].speed;
			}

			reg = ep93xx_calculate_timing(speed);
				reg |= EP93XX_PCMCIA_CONTROLLER_MEMORY_WIDTH_16BIT;

			outl(reg, EP93XX_PCMCIA_CONTROLLER_PCCOMMON); 
			
			skt->speed_mem = speed;
		}
	}

	map->static_start = (map->flags & MAP_ATTRIB) ?
		skt->res_attr->start : skt->res_mem->start;

	skt->mem_map[map->map] = *map;

	return 0;
}


static void init_socket(struct pcmcia_socket *s)
{
	int i;
	pccard_io_map io = {0, 0, 0, 0, 1};
	pccard_mem_map mem = {0, 0, 0, 0, 0, 0};

	mem.static_start = 0;
	s->socket = dead_socket;
	s->ops->set_socket(s, &s->socket);

	for (i = 0; i < EP93XX_MAX_IO_WIN; i++) {
		io.map = i;
		s->ops->set_io_map(s, &io);
		s->io[i].InUse = 0;
		s->io[i].Config = 0;
	}

	for (i = 0; i < EP93XX_MAX_MEM_WIN; i++) {
		mem.map = i;
		s->ops->set_mem_map(s, &mem);
	}
}


static int ep93xx_pcmcia_sock_init(struct pcmcia_socket *sock)
{
	pr_debug(DRIVER_NAME ": ep93xx_pcmcia_sock_init\n");

	init_socket(sock);

	return 0;
}


static int ep93xx_pcmcia_sock_suspend(struct pcmcia_socket *sock)
{
	
	pr_debug(DRIVER_NAME ": ep93xx_pcmcia_sock_suspend - enter\n");

	return 0;
}


static struct pccard_operations ep93xx_pcmcia_operations = {
	.init			= ep93xx_pcmcia_sock_init,
	.suspend		= ep93xx_pcmcia_sock_suspend,
	.get_status		= ep93xx_get_status,
	.set_socket		= ep93xx_set_socket,
	.set_io_map		= ep93xx_set_io_map,
	.set_mem_map		= ep93xx_set_mem_map,
};

static int ep93xx_shutdown(void)
{
	int i;

	pr_debug(DRIVER_NAME ": ep93xx_shutdown - enter\n");

	/* Disable, clear pcmcia irqs in hw */
	ep93xx_disable_irq(EP93XX_GPIOS);

	/* Set reset */
	outl((EP93XX_PCMCIA_CONTROLLER_PCMCIA_CTRL_WEN |
	      EP93XX_PCMCIA_CONTROLLER_PCMCIA_CTRL_PCRST |
	      EP93XX_PCMCIA_CONTROLLER_PCMCIA_CTRL_PCEN),
	     EP93XX_PCMCIA_CONTROLLER_PCMCIA_CTRL);

	/*
	 * Release memory
	 * Free the pcmcia interrupts.
	 * Set socket voltage to zero
	 */
	iounmap(socket_info.virt_io);
	socket_info.vcc = EP93XX_UNCONFIGURED_VOLTAGE;
	ep93xx_set_voltage(0);

	for (i = 0; i < ARRAY_SIZE(ep93xx_pcmcia_irqs) ; i++)
		free_irq(ep93xx_pcmcia_irqs[i].irq, &socket_info);

	return 0;
}

static int ep93xx_drv_pcmcia_probe(struct platform_device *pdev)
{
	struct socket_info_t *skt = &socket_info;
	unsigned long tmp, flags;
	int ret;
	int i;

	pr_debug(DRIVER_NAME ": ep93xx_drv_pcmcia_probe\n");

	skt->res_io = platform_get_resource_byname(pdev, IORESOURCE_MEM,
						   "pcmcia:io");
	skt->res_attr = platform_get_resource_byname(pdev, IORESOURCE_MEM,
						     "pcmcia:attr");
	skt->res_mem = platform_get_resource_byname(pdev, IORESOURCE_MEM,
						    "pcmcia:mem");

	if (!skt->res_io || !skt->res_attr || !skt->res_mem) {
		pr_err(DRIVER_NAME ": Can't find memory resource\n");
		return -ENOENT;
	}

	skt->hclk = clk_get(&pdev->dev, "hclk");

	/* Initialise the socket structure. */
	skt->socket.ops 	= &ep93xx_pcmcia_operations;
	skt->socket.owner 	= THIS_MODULE;
	skt->socket.dev.parent 	= &pdev->dev;
	skt->socket.driver_data = pdev;

	skt->nr  = 0;
	skt->irq = NO_IRQ;
	skt->dev = &pdev->dev;

	for (i = 0; i < EP93XX_MAX_IO_WIN; i++) {
		skt->io_map[i].flags = 0;
		skt->io_map[i].speed = EP93XX_PCMCIA_IO_ACCESS;
	}		

	for (i = 0; i < EP93XX_MAX_MEM_WIN; i++) {
		skt->mem_map[i].flags = 0;
		skt->mem_map[i].speed = EP93XX_PCMCIA_3V_MEM_ACCESS;
	}

	skt->speed_io   = EP93XX_PCMCIA_IO_ACCESS;
	skt->speed_attr = EP93XX_PCMCIA_5V_MEM_ACCESS;
	skt->speed_mem  = EP93XX_PCMCIA_5V_MEM_ACCESS;

	skt->status     = 0;
	skt->enable_irq = 0;
		
	skt->vcc = EP93XX_UNCONFIGURED_VOLTAGE;
	ep93xx_set_voltage(0);

	/* virt memory remap */
	skt->virt_io = ioremap(skt->res_io->start,
			       skt->res_io->end - skt->res_io->start);
	pr_debug(DRIVER_NAME ": virt_io is 0x%p\n", skt->virt_io);
	if (skt->virt_io == NULL) {
		ep93xx_shutdown();
		return -ENOMEM;
	}

	/* Disable interrupts in hw for pcmcia. */
	ep93xx_disable_irq(EP93XX_GPIOS);
		
	/* Enable debounce for the card detect lines */
	tmp = inl(EP93XX_GPIO_F_DEBOUNCE) & ~EP93XX_GPIOS;
	tmp |= EP93XX_PCMCIA_MCD1 | EP93XX_PCMCIA_MCD2; 
	outl(tmp, EP93XX_GPIO_F_DEBOUNCE);

	/* Set interrupts to be edge sensitive, falling edge triggered. */
	for (i = 0; i < ARRAY_SIZE(ep93xx_pcmcia_irqs) ; i++) {
		if (ep93xx_pcmcia_irqs[i].gpio &
		    (EP93XX_PCMCIA_MCD1 | EP93XX_PCMCIA_MCD2))
			set_irq_type(ep93xx_pcmcia_irqs[i].irq, IRQ_TYPE_EDGE_BOTH);
		else
			set_irq_type(ep93xx_pcmcia_irqs[i].irq, IRQ_TYPE_EDGE_FALLING);
	}

	/*
	 * Register ISR. EP93XX_PCMCIA_INT_READY is a shared interrupt as
	 * the kernel IDE stack has its own interrupt handler that it
	 * will register for it.
	 */
	for (i = 0; i < ARRAY_SIZE(ep93xx_pcmcia_irqs); i++) {
		flags = IRQF_DISABLED;
		if (ep93xx_pcmcia_irqs[i].irq == EP93XX_PCMCIA_INT_READY)
			flags |= IRQF_SHARED;

		ret = request_irq(ep93xx_pcmcia_irqs[i].irq,
				  ep93xx_interrupt, flags,
				  ep93xx_pcmcia_irqs[i].str, &socket_info);
		if (ret)
			pr_err(DRIVER_NAME ":failed to request_irq: %d\n",
			       ep93xx_pcmcia_irqs[i].irq);
	}

	/* Set speed to the defaults */
	tmp = ep93xx_calculate_timing(skt->speed_io);
	outl(tmp, EP93XX_PCMCIA_CONTROLLER_PCIO);

	tmp = ep93xx_calculate_timing(skt->speed_attr);
	outl(tmp, EP93XX_PCMCIA_CONTROLLER_PCATTRIBUTE); 

	tmp = ep93xx_calculate_timing(skt->speed_mem);
	outl(tmp, EP93XX_PCMCIA_CONTROLLER_PCCOMMON); 

	pr_debug(DRIVER_NAME ": INITIALIZING SMC: Attr:%08x  Comm:%08x  "
		 "IO:%08x  Ctrl:%08x\n",
		 inl(EP93XX_PCMCIA_CONTROLLER_PCATTRIBUTE),
		 inl(EP93XX_PCMCIA_CONTROLLER_PCCOMMON), 
		 inl(EP93XX_PCMCIA_CONTROLLER_PCIO),
		 inl(EP93XX_PCMCIA_CONTROLLER_PCMCIA_CTRL));

	/*initialize the socket info*/
	skt->socket.features  = (SS_CAP_PAGE_REGS | SS_CAP_VIRTUAL_BUS |
				 SS_CAP_MEM_ALIGN | SS_CAP_STATIC_MAP |
				 SS_CAP_PCCARD);
	skt->socket.irq_mask  = 0;
	skt->socket.map_size  = PAGE_SIZE;
	skt->socket.pci_irq   = EP93XX_PCMCIA_INT_READY;

	skt->socket.sock      = 0;
	skt->socket.io_offset = (unsigned int)skt->virt_io;
	skt->socket.resource_ops = &pccard_static_ops; 
		
	ret = pcmcia_register_socket(&skt->socket);
	if (ret) {
		pr_err(DRIVER_NAME ": Unable to register sockets\n");
		ep93xx_shutdown();
		return ret;
	}

	ep93xx_disable_irq(EP93XX_PCMCIA_READY | EP93XX_PCMCIA_MCD1 |
			   EP93XX_PCMCIA_MCD2);
	ep93xx_enable_irq(EP93XX_PCMCIA_READY | EP93XX_PCMCIA_MCD1 |
			  EP93XX_PCMCIA_MCD2);

	pr_debug(DRIVER_NAME ": ep93xx_init GPIO_F: INTEN=0x%02x  DDR=0x%02x "
		 "DB=0x%02x INTTYP1=0x%02x INTTYP2=0x%02x\n",
		 inl(EP93XX_GPIO_F_INT_ENABLE), inl(EP93XX_GPIO_F_DIRECTION),
		 inl(EP93XX_GPIO_F_DEBOUNCE), inl(EP93XX_GPIO_F_INT_TYPE1),
		 inl(EP93XX_GPIO_F_INT_TYPE2));

	ret = device_create_file(&skt->socket.dev, &dev_attr_status);
	if (ret) {
		pr_err(DRIVER_NAME ": Unable to create sysfs file\n");
		ep93xx_shutdown();
	}

	return ret;
}

static int ep93xx_drv_pcmcia_remove(struct platform_device *pdev)
{
	struct socket_info_t *skt = &socket_info;

	pr_debug(DRIVER_NAME ": ep93xx_remove\n");

	pcmcia_unregister_socket(&skt->socket);

	ep93xx_shutdown();

	return 0;
}

static int
ep93xx_drv_pcmcia_suspend(struct platform_device *pdev, pm_message_t state)
{
	pr_debug(DRIVER_NAME ": ep93xx_suspend\n");

	return pcmcia_socket_dev_suspend(&pdev->dev, state);
}

static int ep93xx_drv_pcmcia_resume(struct platform_device *pdev)
{
	pr_debug(DRIVER_NAME ": ep93xx_resume\n");

	return pcmcia_socket_dev_resume(&pdev->dev);
}

static struct platform_driver ep93xx_pcmcia_platform_driver = {
	.driver		= {
		.name	= DRIVER_NAME,
	},
	.probe		= ep93xx_drv_pcmcia_probe,
	.remove		= ep93xx_drv_pcmcia_remove,
	.suspend	= ep93xx_drv_pcmcia_suspend,
	.resume		= ep93xx_drv_pcmcia_resume,
};

static int __init ep93xx_pcmcia_init(void)
{
	int ret;

	pr_debug(DRIVER_NAME ": ep93xx_pcmcia_init\n");

	ret = platform_driver_register(&ep93xx_pcmcia_platform_driver);
	if (ret)
		pr_err(DRIVER_NAME ": Unable to register.\n");

	return ret;
}

static void __exit ep93xx_pcmcia_exit(void)
{
	pr_debug(DRIVER_NAME ": ep93xx_pcmcia_exit\n");

	platform_driver_unregister(&ep93xx_pcmcia_platform_driver);
}

module_init(ep93xx_pcmcia_init);
module_exit(ep93xx_pcmcia_exit);

MODULE_AUTHOR(AUTHOR);
MODULE_DESCRIPTION(DESCRIPTION);
MODULE_LICENSE("GPL");
